<?php

namespace Drupal\webform_post_remote_queue\Plugin\QueueWorker;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueWorkerBase;
use Drupal\Core\Queue\SuspendQueueException;
use Drupal\Core\Url;
use Drupal\webform\Plugin\WebformElement\BooleanBase;
use Drupal\webform\Plugin\WebformElement\NumericBase;
use Drupal\webform\Plugin\WebformElement\WebformCompositeBase;
use Drupal\webform\Plugin\WebformElement\WebformManagedFileBase;
use Drupal\webform\Plugin\WebformElementInterface;
use Drupal\webform\Plugin\WebformElementManagerInterface;
use Drupal\webform\WebformSubmissionInterface;
use Drupal\webform\WebformTokenManagerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ConnectException;
use Psr\Http\Message\ResponseInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines 'webform_post_remote_queue' queue worker.
 *
 * @QueueWorker(
 *   id = "webform_post_remote_queue",
 *   title = @Translation("Webform Remote Post Queue"),
 *   cron = {"time" = 60}
 * )
 */
class WebformPostRemoteQueue extends QueueWorkerBase implements ContainerFactoryPluginInterface {

  /**
   * The webform token manager.
   *
   * @var \Drupal\webform\WebformTokenManagerInterface
   */
  protected WebformTokenManagerInterface $tokenManager;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The webform element manager.
   *
   * @var \Drupal\webform\Plugin\WebformElementManagerInterface
   */
  protected WebformElementManagerInterface $elementManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The YAML service.
   *
   * @var \Drupal\Component\Serialization\Yaml
   */
  protected Yaml $yaml;

  /**
   * Constructs a new class instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\webform\WebformTokenManagerInterface $token_manager
   *   The webform token manager.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   * @param \Drupal\webform\Plugin\WebformElementManagerInterface $element_manager
   *   The webform element manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   * @param \Drupal\Component\Serialization\Yaml $yaml
   *   The YAML serializer.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, WebformTokenManagerInterface $token_manager, ClientInterface $http_client, WebformElementManagerInterface $element_manager, EntityTypeManagerInterface $entity_type_manager, LoggerInterface $logger, Yaml $yaml) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->tokenManager = $token_manager;
    $this->httpClient = $http_client;
    $this->elementManager = $element_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
    $this->yaml = $yaml;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('webform.token_manager'),
      $container->get('http_client'),
      $container->get('plugin.manager.webform.element'),
      $container->get('entity_type.manager'),
      $container->get('logger.channel.webform'),
      $container->get('serialization.yaml')
    );
  }

  /**
   * List of unsupported webform submission properties.
   *
   * The below properties will not being included in a remote post.
   *
   * @var array
   */
  protected array $unsupportedProperties = [
    'metatag',
  ];

  /**
   * {@inheritdoc}
   */
  public function processItem($data) {
    $state = $data['state'];
    $configuration = $data['configuration'];
    $handler_id = $data['handler_id'];

    /** @var \Drupal\webform\WebformSubmissionInterface $webform_submission */
    $webform_submission = $this->entityTypeManager->getStorage('webform_submission')->load($data['webform_submission_id']);

    // If the webform submission is not found, it may have been deleted.
    if (empty($webform_submission)) {
      return;
    }

    $webform = $webform_submission->getWebform();
    $state_url = $state . '_url';
    if (empty($configuration[$state_url])) {
      // We don't want to requeue this item immediately.
      throw new \Exception();
    }

    $request_url = $configuration[$state_url];
    $request_url = $this->tokenManager->replaceNoRenderContext($request_url, $webform_submission);
    $request_method = (!empty($configuration['method'])) ? $configuration['method'] : 'POST';
    $request_type = ($request_method !== 'GET') ? $configuration['type'] : NULL;

    // Get request options with tokens replaced.
    $request_options = (!empty($configuration['custom_options'])) ? $this->yaml->decode($configuration['custom_options']) : [];
    $request_options = $this->tokenManager->replaceNoRenderContext($request_options, $webform_submission);

    try {
      if ($request_method === 'GET') {
        // Append data as query string to the request URL.
        $query = $this->getRequestData($state, $webform_submission, $configuration);
        $request_url = Url::fromUri($request_url, ['query' => $query])->toString();
        $response = $this->httpClient->request('GET', $request_url, $request_options);
      }
      else {
        $method = strtolower($request_method);
        $request_options[($request_type === 'json' ? 'json' : 'form_params')] = $this->getRequestData($state, $webform_submission, $configuration);
        $response = $this->httpClient->$method($request_url, $request_options);
      }
    }
    catch (ConnectException $e) {
      $this->logger->error('Webform submission %webform_submission_id remote post failed: @message', [
        '%webform_submission_id' => $webform_submission->id(),
        '@message' => $e->getMessage(),
      ]);

      $submission_data = $webform_submission->getData();
      $submission_data['failed'] = TRUE;
      $webform_submission->setData($submission_data);
      if ($webform->getSetting('results_disabled') === FALSE) {
        $webform_submission->resave();
      }
      return;
    }
    catch (\Exception $e) {
      $this->logger->error('Webform submission %webform_submission_id remote post failed: @message', [
        '%webform_submission_id' => $webform_submission->id(),
        '@message' => $e->getMessage(),
      ]);
      // Major issues should stop the queue.
      throw new SuspendQueueException();
    }

    // Display submission exception if response code is not 2xx.
    if (!$response instanceof ResponseInterface || $this->responseHasError($response)) {
      throw new \Exception();
    }
    // Replace [webform:handler] tokens in submission data.
    // Data structured for [webform:handler:remote_post:completed:key] tokens.
    $submission_data = $webform_submission->getData();
    $submission_has_token = (strpos(print_r($submission_data, TRUE), '[webform:handler:' . $handler_id . ':') !== FALSE) ? TRUE : FALSE;
    if ($submission_has_token) {
      $response_data = $this->getResponseData($response);
      $token_data = ['webform_handler' => [$handler_id => [$state => $response_data]]];
      $submission_data = $this->tokenManager->replaceNoRenderContext($submission_data, $webform_submission, $token_data);
      $webform_submission->setData($submission_data);
      // Resave changes to the submission data without invoking any hooks
      // or handlers.
      if ($webform->getSetting('results_disabled') === FALSE) {
        $webform_submission->resave();
      }
    }
  }

  /**
   * Get a webform submission's request data.
   *
   * @param string $state
   *   The state of the webform submission.
   *   Either STATE_NEW, STATE_DRAFT_CREATED, STATE_DRAFT_UPDATED,
   *   STATE_COMPLETED, STATE_UPDATED, or STATE_CONVERTED
   *   depending on the last save operation performed.
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission to be posted.
   * @param array $configuration
   *   Handler configuration array.
   *
   * @return array
   *   A webform submission converted to an associative array.
   *
   * @throws \Exception
   */
  protected function getRequestData($state, WebformSubmissionInterface $webform_submission, array $configuration) {
    // Get submission and elements data.
    $data = $webform_submission->toArray(TRUE);

    // Remove unsupported properties from data.
    // These are typically added by other module's like metatag.
    $unsupported_properties = array_combine($this->unsupportedProperties, $this->unsupportedProperties);
    $data = array_diff_key($data, $unsupported_properties);

    // Flatten data and prioritize the element data over the
    // webform submission data.
    $element_data = $data['data'];
    unset($data['data']);
    $data = $element_data + $data;

    // Excluded selected submission data.
    $data = array_diff_key($data, $configuration['excluded_data']);

    // Append uploaded file name, uri, and base64 data to data.
    $webform = $webform_submission->getWebform();
    foreach ($data as $element_key => $element_value) {
      // Ignore empty and not equal to zero values.
      // @see https://stackoverflow.com/questions/732979/php-whats-an-alternative-to-empty-where-string-0-is-not-treated-as-empty
      if (empty($element_value) && $element_value !== 0 && $element_value !== '0') {
        continue;
      }

      $element = $webform->getElement($element_key);
      if (!$element) {
        continue;
      }

      // Cast markup to string. This only applies to computed Twig values.
      // @see \Drupal\webform\Element\WebformComputedTwig::computeValue
      if ($element_value instanceof MarkupInterface) {
        $data[$element_key] = $element_value = (string) $element_value;
      }

      $element_plugin = $this->elementManager->getElementInstance($element);

      if ($element_plugin instanceof WebformManagedFileBase) {
        if ($element_plugin->hasMultipleValues($element)) {
          foreach ($element_value as $fid) {
            $data['_' . $element_key][] = $this->getRequestFileData($fid, $configuration);
          }
        }
        else {
          $data['_' . $element_key] = $this->getRequestFileData($element_value, $configuration);
        }
      }
      elseif (!empty($configuration['cast'])) {
        // Cast value.
        $data[$element_key] = $this->castRequestValues($element, $element_plugin, $element_value);
      }
    }

    // Replace tokens.
    $data = $this->tokenManager->replaceNoRenderContext($data, $webform_submission);

    // Append custom data.
    if (!empty($configuration['custom_data'])) {
      $custom_data = $this->yaml->decode($configuration['custom_data']);
      // Replace tokens.
      $custom_data = $this->tokenManager->replaceNoRenderContext($custom_data, $webform_submission);
      // Cast custom data.
      $custom_data = $this->tokenManager->replaceNoRenderContext($custom_data);
      $data = $custom_data + $data;
    }

    // Append state custom data.
    if (!empty($configuration[$state . '_custom_data'])) {
      $state_custom_data = $this->yaml->decode($configuration[$state . '_custom_data']);
      // Replace tokens.
      $state_custom_data = $this->tokenManager->replaceNoRenderContext($state_custom_data, $webform_submission);
      // Cast custom data.
      $state_custom_data = $this->castCustomData($state_custom_data, $configuration);
      $data = $state_custom_data + $data;
    }

    return $data;
  }

  /**
   * Cast custom data.
   *
   * @param array $data
   *   Custom data.
   * @param array $configuration
   *   Handler configuration array.
   *
   * @return array
   *   The custom data with value casted
   */
  protected function castCustomData(array $data, array $configuration) {
    if (empty($configuration['cast'])) {
      return $data;
    }

    foreach ($data as $key => $value) {
      if (is_array($value)) {
        $data[$key] = $this->castCustomData($value, $configuration);
      }
      elseif (is_string($value) && preg_match('/^\((int|integer|bool|boolean|float|double|real)\)\s*(.+)$/', $value, $match)) {
        $type_cast = $match[1];
        $type_value = $match[2];
        switch ($type_cast) {
          case 'int':
          case 'integer':
            $data[$key] = (int) $type_value;
            break;

          case 'bool':
          case 'boolean';
            $data[$key] = (bool) $type_value;
            break;

          case 'float':
          case 'double':
          case 'real':
            $data[$key] = (float) $type_value;
            break;
        }
      }
    }
    return $data;
  }

  /**
   * Cast request values.
   *
   * @param array $element
   *   An element.
   * @param \Drupal\webform\Plugin\WebformElementInterface $element_plugin
   *   The element's webform plugin.
   * @param mixed $value
   *   The element's value.
   *
   * @return mixed
   *   The element's values cast to boolean or float when appropriate.
   */
  protected function castRequestValues(array $element, WebformElementInterface $element_plugin, $value) {
    $element_plugin->initialize($element);
    if ($element_plugin->hasMultipleValues($element)) {
      foreach ($value as $index => $item) {
        $value[$index] = $this->castRequestValue($element, $element_plugin, $item);
      }
      return $value;
    }
    else {
      return $this->castRequestValue($element, $element_plugin, $value);
    }
  }

  /**
   * Cast request value.
   *
   * @param array $element
   *   An element.
   * @param \Drupal\webform\Plugin\WebformElementInterface $element_plugin
   *   The element's webform plugin.
   * @param mixed $value
   *   The element's value.
   *
   * @return mixed
   *   The element's value cast to boolean or float when appropriate.
   */
  protected function castRequestValue(array $element, WebformElementInterface $element_plugin, $value) {
    if ($element_plugin instanceof BooleanBase) {
      return (boolean) $value;
    }
    elseif ($element_plugin instanceof NumericBase) {
      return (float) $value;
    }
    elseif ($element_plugin instanceof WebformCompositeBase) {
      $composite_elements = (isset($element['#element']))
        ? $element['#element']
        : $element_plugin->getCompositeElements();
      foreach ($composite_elements as $key => $composite_element) {
        if (isset($value[$key])) {
          $composite_element_plugin = $this->elementManager->getElementInstance($composite_element);
          $value[$key] = $this->castRequestValue($composite_element, $composite_element_plugin, $value[$key]);
        }
      }
      return $value;
    }
    else {
      return $value;
    }
  }

  /**
   * Get request file data.
   *
   * @param int $fid
   *   A file id.
   * @param array $configuration
   *   Handler configuration array.
   * @param string|null $prefix
   *   A prefix to prepended to data.
   *
   * @return array
   *   An associative array containing file data (name, uri, mime, and data).
   */
  protected function getRequestFileData($fid, array $configuration, $prefix = '') {
    /** @var \Drupal\file\FileInterface $file */
    $file = $this->entityTypeManager->getStorage('file')->load($fid);
    if ($file == NULL) {
      return [];
    }

    $data = [];
    $data[$prefix . 'id'] = (int) $file->id();
    $data[$prefix . 'name'] = $file->getFilename();
    $data[$prefix . 'uri'] = $file->getFileUri();
    $data[$prefix . 'mime'] = $file->getMimeType();
    $data[$prefix . 'uuid'] = $file->uuid();
    if ($configuration['file_data']) {
      $data[$prefix . 'data'] = base64_encode(file_get_contents($file->getFileUri()));
    }
    return $data;
  }

  /**
   * Determine if response has an error status code.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   The response returned by the remote server.
   *
   * @return bool
   *   TRUE if response status code reflects an unsuccessful value.
   */
  protected function responseHasError($response) {
    $status_code = $response->getStatusCode();
    return $status_code < 200 || $status_code >= 300;
  }

  /**
   * Get response data.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   The response returned by the remote server.
   *
   * @return array|string
   *   An array of data, parse from JSON, or a string.
   */
  protected function getResponseData(ResponseInterface $response) {
    $body = (string) $response->getBody();
    $data = json_decode($body, TRUE);
    return (json_last_error() === JSON_ERROR_NONE) ? $data : $body;
  }

}
